#include <stdio.h>
#include "twcommon.h"
#include "twsys.h"
#include "tester.h"
#include "machdep.h"
#include "pvmuser.h"

extern char pvm_name[];
extern int node_add[];

/*
#define JOURNAL_FILE	1
#define MSGTRACE
*/

#if JOURNAL_FILE
static FILE *fp;
static char filename[35];
#endif

/* 
 *	note that these routines will not work in the standalone version
 *
 */

get_msg ( rs )
MSG_STRUCT * rs;
{
	register LowLevelMsgH * ll_msg = (LowLevelMsgH *) (rs->buf);
	register char * msgbuf = (char *) ll_msg;
	register int node;
	int inum;


#if JOURNAL_FILE
	if ( !fp )
	{
		sprintf ( filename, "communications%d", tw_node_num );
		fp = fopen ( filename, "w" );
	}
#endif

	while ( probe ( -1 ) != PvmNoMsg )
	{
		rs->type = rcv ( -1 );

		getnint ( &(rs->source), 1 );
		getnint ( &(rs->dest), 1 );
		getnint ( &(rs->type), 1 );
		getnint ( &(rs->mlen), 1 );
		getbytes ( msgbuf, rs->mlen );

		rcvinfo ( NULL, NULL, NULL, &inum );

		node = rs->source;

#ifdef MSGTRACE
		printf ( "recv type %d, len %d from instance %d\n",
			rs->type, rs->mlen, inum ); 
#endif


#if JOURNAL_FILE
		fprintf ( fp, "get_msg header src %d, dest %d, type %d, len %d\n",
			rs->source, rs->dest, rs->type, rs->mlen				);

		dumpit ( msgbuf, rs->mlen );
#endif

		if ( rs->type == ACK_MSG )   /* ACK */
		{
			rcvack ( ll_msg, node );
			continue;
		}

#if CHECKSUM
		if ( checksum ( msgbuf ) != ((Msgh *)msgbuf)->checksum )
		{
			_pprintf ( "Checksum Error\n" );
			showmsg ( msgbuf );
			tester ();
		}
#endif
		if ( (node != tw_node_num) && (node != CP)
			&&  !(((Msgh *)msgbuf)->flags & (SYSMSG|MOVING))
			&& (strcmp ( ((Msgh *)msgbuf)->rcver, "$IH" ) != 0) )
		{
			sndack ( msgbuf, node );
		}

		if ( (node == CP) && ( rs->type == TESTER_COMMAND ) )
		{
			extern char buff[], * bp;
			extern int host_input_waiting;

			msgbuf[rs->mlen] = '\0';

			strcpy ( buff, msgbuf );
			bp = buff;
			host_input_waiting = TRUE;

			return FALSE;
		}
		return TRUE;
	}
	return FALSE;
}

get_msg_w ( rs )
MSG_STRUCT    *rs;
{
	for (;;)
	{
		if ( get_msg ( rs ) )
			return TRUE;
	}
}

send_msg ( ss )
MSG_STRUCT  *ss;
{
	register LowLevelMsgH * ll_msg = (LowLevelMsgH *) (ss->buf);
	register char * msgbuf = (char *) ll_msg;
	register int ret;

	extern int mlog, node_cputime;

#if JOURNAL_FILE
	if ( !fp )
	{
		sprintf ( filename, "communications%d", tw_node_num );
		fp = fopen ( filename, "w" );
	}
#endif


#ifdef MSGTRACE
		printf ( "send type %d, len %d to node %d\n",
			ss->type, ss->mlen, ss->dest ); 
#endif

#if PARANOID
	/* this dies for non-normal msgs */
	if ( ( ss->mlen != ll_msg->length ) ||
		( ss->dest != ll_msg->to_node ) ||
		( ss->source != ll_msg->from_node ) ||
		( ss->type != ll_msg->type ) )

		if ( ll_msg->to_node != IH_NODE )
			printf ( "parameter missmatch in send_msg trying to send to %d\n",
				ll_msg->to_node );
#endif
		
#if JOURNAL_FILE
	fprintf ( fp, "send_msg header src %d, dest %d, type %d, len %d\n",
		ss->source, ss->dest, ss->type, ss->mlen );
	dumpit ( msgbuf, ss->mlen);
#endif

/*
	if ( ss->type == ACK_MSG )
	{
		printf ( "send_msg: given ACK\n" );
		tester();
	}
*/
	if ( ss->dest == ALL )
	{
		printf ( "SUN send_msg doesn't b'cst\n" );
		tester();
	}

	if ( ss->dest == tw_node_num )
	{
		printf ( "send_msg doesn't send to self add: 0x%x\n", ll_msg );
		tester();
	}

#if MICROTIME
	if ( mlog && (ss->type != ACK_MSG) && ( ss->dest != CP ) )
	{
		MicroTime ();
		((Msgh *)msgbuf)->msgtimef = node_cputime;
	}
#endif

	pvm_inithow ( PvmDataRaw ); /* no translation */

	putnint ( &(ss->source), 1 );
	putnint ( &(ss->dest), 1 );
	putnint ( &(ss->type), 1 );
	putnint ( &(ss->mlen), 1 );
	putbytes ( msgbuf, ss->mlen );

	ret = ss->mlen;

/*
	printf ( "snd( %s %d %d )\n", pvm_name, node_add[ss->dest], ss->type );
*/
	if ( snd ( pvm_name, node_add[ss->dest], ss->type ) < 0 )
	{
		/* failure */
		printf ( "pvm send failed\n" );
		ret = -1;
	}

#if JOURNAL_FILE
	fprintf ( fp, "send_msg returning %d\n\n", ret );
#endif

	return ( ret ); /* -1 indicates failure */
}

send_msg_w ( ss )
MSG_STRUCT  *ss;
{
	register int ret;
	register int count = 0;

	while ( -1 == (ret = send_msg ( ss ) ) )
	{
		if ( count++ >= 1000 )
		{
			count = 0;
			_pprintf ( "Node %d stuck writing to node %d\n", tw_node_num,
				ss->dest );
		}
	}

	return ( ret ); /* -1 indicates failure */
}

show_ll_msgh ( sb )
LowLevelMsgH * sb;
{
  printf ("length =%d to_node =%d from_node =%d type =%d\n",
			sb->length, sb->from_node, sb->to_node, sb->type );
}

extern char buff[], * bp;
extern int host_input_waiting;

get_tester_input ()
{
	int src, dest, type, mlen, inum;

	rcv ( TESTER_COMMAND ); /* this blocks */

	getnint ( &src, 1 );
	getnint ( &dest, 1 );
	getnint ( &type, 1 );
	getnint ( &mlen, 1 );

	getbytes ( buff, mlen );

	rcvinfo ( NULL, NULL, NULL, &inum );

#ifdef MSGTRACE
	printf ( "recv type %d, len %d from instance %d\n",
		type, mlen, inum ); 
#endif

	if ( type != TESTER_COMMAND )
	{
		printf ("get_tester_input got wrong type %d\n", type );
	}

	if ( mlen >= 512 )
	{
		printf ( "Tester input too long! overwriten buff\n" );
	}

	buff[mlen] = '\0';
	bp = buff;

	return TRUE;
}

string_to_cp_w ( string )
char * string;
{
	MSG_STRUCT m;
	register int ret;
	register int count = 0;


	m.mlen = strlen ( string ) + 1;
	m.dest = CP;
	m.source = tw_node_num;
	m.type = TESTER_COMMAND;
	m.buf = (int *)string;

	while ( -1 == (ret = send_msg ( &m ) ) )
	{
		if ( count++ >= 1000 )
		{
			count = 0;
			printf ( "Node %d Stuck writing string to CP\n", tw_node_num );
		}
	}

	return ( ret ); /* -1 indicates failure */
}

#if JOURNAL_FILE
dumpit (u, n)
	char           *u;
	int             n;
{
	char            s[80];
	char            t[10],
					c;
	int             i,
					j;

	s[78] = '\n';
	s[79] = '\0';

#define clrs    {int i; for(i=0; i<78; i++) s[i] = ' ';}
#define toupper(c)	(( c >= 'a' && c <= 'z' )? c - ' ': c)

	for (i = 0; i < n;)
	{
		clrs

		sprintf (t, "%3.3x:", i % 0x1000);

		s[0] = toupper (t[0]);
		s[1] = toupper (t[1]);
		s[2] = toupper (t[2]);
		s[3] = t[3];

		for (j = 0; j < 16 && i < n; j++, i++)
		{
			sprintf (t, "%2x ", 0377 & (c = 0377 & u[i]));
			s[5 + 3 * j] = toupper (t[0]);
			s[5 + 3 * j + 1] = toupper (t[1]);
			s[5 + 52 + j] = c < 32 || c > 127 ? '.' : c;
		}
		fprintf (fp, "%s", s);
	}
		fprintf (fp, "\n" );
		fflush (fp);
}
#endif
